import pytest
from django import forms as django_forms

from reactivated import serialization
from reactivated.forms import FormGroup
from sample.server.apps.samples import forms, models

from .serialization import convert_to_json_and_validate


class ComposerForm(django_forms.ModelForm):
    class Meta:
        model = models.Composer
        fields = ["name"]


ComposerFormSet = django_forms.formset_factory(ComposerForm)


class OperaFormGroup(FormGroup):
    opera: forms.OperaForm
    composer: ComposerForm
    composers: ComposerFormSet


@pytest.mark.django_db
def test_form_group():
    opera_form = forms.OperaForm({})
    composer_form = ComposerForm({})
    form_group = OperaFormGroup({})

    assert form_group.is_valid() is False
    assert form_group.errors["opera"] == opera_form.errors
    assert form_group.errors["composer"] == composer_form.errors

    composer = models.Composer.objects.create(name="Richard Wagner")

    form_group = OperaFormGroup(
        {
            "opera-name": "test",
            "opera-composer": composer.id,
            "opera-style": models.Opera.Style.GRAND,
            "composer-name": "John",
            "composers-0-name": "John",
            "composers-INITIAL_FORMS": 1,
            "composers-TOTAL_FORMS": 1,
        }
    )
    assert form_group.is_valid() is True


@pytest.mark.django_db
def test_form_group_instance(snapshot):
    generated_schema = serialization.create_schema(OperaFormGroup, {})
    serialized = serialization.serialize(OperaFormGroup(), generated_schema)
    convert_to_json_and_validate(serialized, generated_schema)
    assert generated_schema == snapshot
    assert serialized == snapshot


def test_form_group_class_not_instance(snapshot):
    generated_schema = serialization.create_schema(OperaFormGroup, {})
    serialized = serialization.serialize(OperaFormGroup, generated_schema)
    convert_to_json_and_validate(serialized, generated_schema)
    assert generated_schema == snapshot
    assert serialized == snapshot


def test_form_class_not_instance(snapshot):
    generated_schema = serialization.create_schema(forms.OperaForm, {})
    serialized = serialization.serialize(forms.OperaForm, generated_schema)
    assert generated_schema == snapshot
    assert serialized == snapshot


def test_form_set_class_not_instance(snapshot):
    generated_schema = serialization.create_schema(forms.OperaFormSet, {})
    serialized = serialization.serialize(forms.OperaFormSet, generated_schema)
    assert generated_schema == snapshot
    assert serialized == snapshot


class FormWithModel(django_forms.Form):
    model_choice = django_forms.ModelChoiceField(queryset=models.Opera.objects.all())


def test_model_choice_fields_without_db_access_failure(snapshot):
    generated_schema = serialization.create_schema(FormWithModel, {})
    with pytest.raises(BaseException, match="Database access not allowed"):
        serialization.serialize(FormWithModel, generated_schema)


def test_model_choice_fields_without_db_access_passing(snapshot):
    generated_schema = serialization.create_schema(FormWithModel, {})
    generated_schema.definitions["is_static_context"] = True
    serialization.serialize(FormWithModel, generated_schema)



from __future__ import annotations

import collections
import inspect
from typing import TYPE_CHECKING, Any, Sequence

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Manager
from django.utils.functional import Promise

if TYPE_CHECKING:
    from reactivated.backend import JSX


# Mock is_simple_callable for now
# instead of the more sophisticated one from rest_framework.fields
def is_simple_callable(possible_callable: Any) -> bool:
    return callable(possible_callable)


# From Django Rest Framework 3.6
def get_attribute(instance: Any, attrs: Sequence[str]) -> Any:
    """
    Similar to Python's built in `getattr(instance, attr)`,
    but takes a list of nested attributes, instead of a single attribute.
    Also accepts either attribute lookup on objects or dictionary lookups.
    """
    for attr in attrs:
        if instance is None:
            # Break out early if we get `None` at any point in a nested lookup.
            return None
        try:
            if isinstance(instance, collections.Mapping):
                instance = instance[attr]
            else:
                instance = getattr(instance, attr)
        except ObjectDoesNotExist:
            return None

        if isinstance(instance, Manager):
            instance = instance.all()
        elif is_simple_callable(instance):
            try:
                instance = instance()
            except (AttributeError, KeyError) as exc:
                # If we raised an Attribute or KeyError here it'd get treated
                # as an omitted field in `Field.get_attribute()`. Instead we
                # raise a ValueError to ensure the exception is not masked.
                raise ValueError(
                    'Exception raised in callable attribute "{0}"; original exception was: {1}'.format(
                        attr, exc
                    )
                )

    return instance


# From https://github.com/encode/django-rest-framework/blob/master/rest_framework/utils/field_mapping.py
class ClassLookupDict:
    """
    Takes a dictionary with classes as keys.
    Lookups against this object will traverses the object's inheritance
    hierarchy in method resolution order, and returns the first matching value
    from the dictionary or raises a KeyError if nothing matches.
    """

    def __init__(self, mapping: Any) -> None:
        self.mapping = mapping

    def __getitem__(self, key: Any) -> Any:
        base_class = key

        # Handle lazy translated strings, as could come from say labels,
        # verbose names, etc.

        # Useful when a Union is typed as Union[str, somethingelse] but
        # actually receives a proxy at runtime.
        if issubclass(base_class, Promise):
            if getattr(base_class, "_delegate_text", False):
                base_class = str
            else:
                assert (
                    False
                ), "Unsupported proxy / lazy promise type. See django/utils/functional.py"

        # DRF uses this, we do not.
        # if hasattr(key, "_proxy_class"):
        #     # Deal with proxy classes. Ie. BoundField behaves as if it
        #     # is a Field instance when using ClassLookupDict.
        #     base_class = key._proxy_class
        # else:
        #     base_class = key.__class__

        for cls in inspect.getmro(base_class):
            if cls in self.mapping:
                return self.mapping[cls]
        raise KeyError("Class %s not found in lookup." % base_class.__name__)

    def __setitem__(self, key: Any, value: Any) -> Any:
        self.mapping[key] = value


def get_template_engine() -> JSX:
    from django.template import engines

    from reactivated.backend import JSX

    for engine in engines.all():
        if isinstance(engine, JSX):
            return engine
    assert False, "JSX engine not found in settings.TEMPLATES"



import enum
from typing import Dict, Literal, Union, cast

import pytest
from django import forms
from django.http import HttpRequest, HttpResponseRedirect
from django.urls import reverse

from reactivated import Pick
from reactivated.rpc import EmptyForm, FormGroup, RPCRequest, RPCResponse, create_rpc
from sample.server.apps.samples import models

urlpatterns = []


class SimpleForm(forms.Form):
    char_field = forms.CharField()


class SpecializedHttpRequest(HttpRequest):
    extra_property: bool


special_rpc = create_rpc(
    lambda request: cast(SpecializedHttpRequest, request)
    if request.user.is_authenticated
    else False
)

anonymous_rpc = create_rpc(lambda request: request)


@anonymous_rpc
def simple_form(request: HttpRequest, form: SimpleForm) -> bool:
    return True


urlpatterns.append(simple_form)


@pytest.mark.urls("tests.rpc")
def test_simple_form(client):
    url = reverse("rpc_simple_form")
    response = client.get(url)
    assert response.status_code == 405

    response = client.post(url)

    assert response.status_code == 400
    assert response.json() == {"char_field": ["This field is required."]}

    response = client.post(url, {"char_field": "content"})
    assert response.status_code == 200
    assert response.json() is True


@anonymous_rpc
def non_ajax_redirect(request: HttpRequest, form: EmptyForm) -> RPCResponse[bool]:
    if True:
        return HttpResponseRedirect("/")
    return False


urlpatterns.append(non_ajax_redirect)


@pytest.mark.urls("tests.rpc")
def test_non_ajax_redirect(client):
    url = reverse("rpc_non_ajax_redirect")

    response = client.post(url)
    assert isinstance(response, HttpResponseRedirect)


@anonymous_rpc
def empty_form(request: HttpRequest, form: None) -> bool:
    return True


urlpatterns.append(empty_form)


@pytest.mark.urls("tests.rpc")
def test_empty_form(client):
    url = reverse("rpc_empty_form")

    response = client.get(url)
    assert response.status_code == 405

    response = client.post(url)

    assert response.status_code == 200
    assert response.json() is True


class FormOne(forms.Form):
    field = forms.CharField()


class FormTwo(forms.Form):
    another_field = forms.CharField()


class SimpleFormGroup(FormGroup):
    first: FormOne
    second: FormTwo


@anonymous_rpc
def form_group(request: HttpRequest, form: SimpleFormGroup) -> bool:
    return True


urlpatterns.append(form_group)


@pytest.mark.urls("tests.rpc")
def test_form_group(client):
    url = reverse("rpc_form_group")

    response = client.get(url)
    assert response.status_code == 405

    response = client.post(url, {})
    assert response.status_code == 400
    assert response.json() == {
        "first": {"field": ["This field is required."]},
        "second": {"another_field": ["This field is required."]},
    }

    response = client.post(
        url, {"first-field": "Hello", "second-another_field": "Goodbye"}
    )
    assert response.status_code == 200
    assert response.json() is True


@anonymous_rpc.context
def opera(request: HttpRequest, pk: int) -> models.Opera:
    return models.Opera(pk=pk, name="My Opera Context")


class OperaForm(forms.ModelForm[models.Opera]):
    class Meta:
        model = models.Opera
        fields = ["name"]


@opera.process
def update_opera(
    request: HttpRequest, opera: models.Opera, form: OperaForm
) -> Dict[str, str]:
    return {
        "from_context": opera.pk,
        "from_form": form.cleaned_data["name"],
    }


urlpatterns.append(update_opera)


@pytest.mark.urls("tests.rpc")
def test_context_model_form(client):
    url = reverse("rpc_update_opera", args=[923])

    response = client.get(url)
    assert response.status_code == 405

    response = client.post(url, {})
    assert response.status_code == 400

    response = client.post(url, {"name": "Hello"})
    assert response.status_code == 200
    assert response.json() == {
        "from_context": "923",
        "from_form": "Hello",
    }


@opera.process
def validation_error(
    request: HttpRequest, opera: models.Opera, form: OperaForm
) -> None:
    if form.cleaned_data["name"] == "trigger field error":
        raise forms.ValidationError({"name": "field error"})

    if form.cleaned_data["name"] == "trigger non field error":
        raise forms.ValidationError("non field error")
    # raise forms.ValidationError("Non field error")


urlpatterns.append(validation_error)


@pytest.mark.urls("tests.rpc")
def test_rpc_validation_error(client):
    url = reverse("rpc_validation_error", args=[923])

    response = client.post(url, {"name": "trigger field error"})
    assert response.status_code == 400
    assert response.json()["name"] == ["field error"]

    response = client.post(url, {"name": "trigger non field error"})
    assert response.status_code == 400
    assert response.json()["__all__"] == ["non field error"]


@opera.process
def update_opera_with_no_form(
    request: HttpRequest, opera: models.Opera, form: None
) -> bool:
    return True


urlpatterns.append(update_opera_with_no_form)


@pytest.mark.urls("tests.rpc")
def test_update_opera_with_no_form(client):
    url = reverse("rpc_update_opera_with_no_form", args=[923])

    response = client.get(url)
    assert response.status_code == 405

    response = client.post(url)
    assert response.status_code == 200
    assert response.json() is True


OperaSchema = Pick[models.Opera, "id", "name"]


@opera.rpc
def opera_detail(request: HttpRequest, opera: models.Opera) -> OperaSchema:
    return opera


urlpatterns.append(opera_detail)


@pytest.mark.urls("tests.rpc")
def test_opera_detail(client):
    url = reverse("rpc_opera_detail", args=[923])
    expected = {
        "name": "My Opera Context",
        "id": 923,
    }

    response = client.get(url)
    assert response.status_code == 200
    assert response.json() == expected

    response = client.post(url)
    assert response.status_code == 200
    assert response.json() == expected


class Permissions(enum.Enum):
    BECAUSE_I_SAID_SO = "BECAUSE_I_SAID_SO"


@create_rpc
def denied_rpc(request: HttpRequest) -> RPCRequest[HttpRequest, Permissions]:
    return Permissions.BECAUSE_I_SAID_SO


@denied_rpc
def denied(request: HttpRequest, form: None) -> bool:
    return True


urlpatterns.append(denied)


@pytest.mark.urls("tests.rpc")
def test_denied(client):
    url = reverse("rpc_denied")

    response = client.post(url)
    assert response.status_code == 403
    assert response.json() == "BECAUSE_I_SAID_SO"


@create_rpc
def unauthorized_rpc(request: HttpRequest) -> Union[HttpRequest, Literal[False]]:
    return False


@unauthorized_rpc
def unauthorized(request: HttpRequest, form: None) -> bool:
    return True


urlpatterns.append(unauthorized)


@pytest.mark.urls("tests.rpc")
def test_unauthorized(client):
    url = reverse("rpc_unauthorized")

    response = client.post(url)
    assert response.status_code == 401
    assert response.json() == "UNAUTHORIZED"
